;; Generated package description from vertico.el  -*- no-byte-compile: t -*-
(define-package "vertico" "1.7" "VERTical Interactive COmpletion" '((emacs "27.1") (compat "29.1.4.4")) :commit "cf8b2abf5207696c054c33214c86e3969d415054" :authors '(("Daniel Mendler" . "mail@daniel-mendler.de")) :maintainer '("Daniel Mendler" . "mail@daniel-mendler.de") :keywords '("convenience" "files" "matching" "completion") :url "https://github.com/minad/vertico")
